sjcl.ecc.ecdsa.secretKey.prototype.signDeterministically = function(hash, hashFn) {
  hashFn = hashFn || sjcl.hash.sha512;

  var R  = this._curve.r,
      l  = R.bitLength(),
      lB = Math.ceil(l/8);

  // Deterministic ECDSA based on RFC 6979
  // http://tools.ietf.org/html/rfc6979

  // Set:
  //
  //    V = 0x01 0x01 0x01 ... 0x01
  //
  // such that the length of V, in bits, is equal to 8*ceil(hlen/8).
  var V = new Array(Math.ceil(l/8)).map(function () { return 1; });

  // Set:
  //
  //    K = 0x00 0x00 0x00 ... 0x00
  //
  // such that the length of K, in bits, is equal to 8*ceil(hlen/8).
  var K = new Array(Math.ceil(l/8)).map(function () { return 0; });

  // Set:
  //
  //    K = HMAC_K(V || 0x00 || int2octets(x) || bits2octets(h1))
  var HMAC_K = new sjcl.misc.hmac(sjcl.codec.bytes.toBits(K), hashFn);
  K = HMAC_K.encrypt(sjcl.bitArray.concat(
    sjcl.codec.bytes.toBits(V),
    sjcl.codec.bytes.toBits([0x00]),
    this._exponent.toBits(),
    hash
  ));

  // Set:
  //
  //    V = HMAC_K(V)
  HMAC_K = new sjcl.misc.hmac(sjcl.codec.bytes.toBits(K), hashFn);
  V = HMAC_K.encrypt(V);

  // Set:
  //
  //    K = HMAC_K(V || 0x01 || int2octets(x) || bits2octets(h1))
  //
  // Note that the "internal octet" is 0x01 this time.
  HMAC_K = new sjcl.misc.hmac(sjcl.codec.bytes.toBits(K), hashFn);
  K = HMAC_K.encrypt(sjcl.bitArray.concat(
    sjcl.codec.bytes.toBits(V),
    sjcl.codec.bytes.toBits([0x01]),
    this._exponent.toBits(),
    hash
  ));

  // Set:
  //
  //    V = HMAC_K(V)
  HMAC_K = new sjcl.misc.hmac(sjcl.codec.bytes.toBits(K), hashFn);
  V = HMAC_K.encrypt(V);

  var k, r, validK;
  do {
    // Set T to the empty sequence. The length of T (in bits) is
    // denoted tlen; thus, at that point, tlen = 0.
    var T = [], tlen = 0;

    // While tlen < qlen, do the following:
    while (tlen < l) {
      // V = HMAC_K(V)
      V = HMAC_K.encrypt(V);

      // T = T || V
      T = T.concat(sjcl.codec.bytes.fromBits(V));
      tlen = T.length * 8;
    }

    // Compute:
    //
    //    k = bits2int(T)
    k = sjcl.bn.fromBits(sjcl.bitArray.clamp(sjcl.codec.bytes.toBits(T), l));
    r = this._curve.G.mult(k).x.mod(R);

    // If that value of k is within the [1,q-1] range, and is
    // suitable for DSA or ECDSA (i.e., it results in an r value
    // that is not 0; see Section 3.4), then the generation of k is
    // finished.  The obtained value of k is used in DSA or ECDSA.
    validK = k.equal(sjcl.bn(0)) || k.greaterEquals(R) || r.equal(sjcl.bn(0));
    if (!validK) {
      // Otherwise, compute:
      //
      //         K = HMAC_K(V || 0x00)
      //
      //         V = HMAC_K(V)
      K = HMAC_K.encrypt(sjcl.bitArray.concat(
        sjcl.codec.bytes.toBits(V),
        sjcl.codec.bytes.toBits([0x00])
      ));
      HMAC_K = new sjcl.misc.hmac(sjcl.codec.bytes.toBits(K), hashFn);
      V = HMAC_K.encrypt(V);
    }
  } while (!validK);

  // We're using the parameter "fixedKForTesting" to pass in our strengthened
  // value for k. Based on that name it doesn't sound like a very stable
  // interface. Fortunately, we have SJCL imported into our repository and we
  // have unit tests covering our back, so we should be ok.
  return this.sign(hash, 0, false, k);
};

